import os
import sys
forge = sys.modules["forge"]

# NOTE: The Houdini version of boost may be different.
# The boost-regex appears to be sensitive and they don't provide the dso.
# Moving aside their include/boost seems to fix regex,
# but it leaves their other boost dso's cross-versioned with system includes.

# NOTE: Houdini ode includes also conflict
# Try moving aside their include/ode.

def setup(module):
	srcList = [	"houdini.pmh",
				"HoudiniDraw",
				"HoudiniSOP",
				"SurfaceAccessibleHoudini",
				"SurfaceAccessorHoudini",
				"SurfaceTrianglesHoudini",
				"houdiniDL" ]

	# force all of FE to use Houdini's version of boost
#	forge.includemap['houdini'] = os.path.join(os.getenv('HT'),'include')

	dll = module.DLL( "fexHoudiniDL", srcList )

	# this overrides previous -W flags in codegen
	module.cppmap = {}
	module.cppmap['houdini'] = ' -Wno-deprecated -Wall -W -Wno-parentheses -Wno-sign-compare -Wno-reorder -Wno-uninitialized -Wunused -Wno-unused-parameter'

	deplibs = forge.corelibs+ [
				"fexDrawDLLib",
				"fexOperatorDLLib",
				"fexSurfaceDLLib"	]

	dll.linkmap = {}
	if forge.fe_os == "FE_LINUX":
		# needed for gcc 4.1, not for gcc 4.4, but not sure otherwise
		if forge.compiler_version <= 4.1:
			module.cppmap['houdini'] += ' -DNEED_SPECIALIZATION_STORAGE'

		dll.linkmap["houdinilibs"] = "-shared"
		dll.linkmap["houdinilibs"] += " -L"+os.getenv('HDSO')
		dll.linkmap["houdinilibs"] += " -lHoudiniUI -lHoudiniOPZ -lHoudiniOP3 -lHoudiniOP2 -lHoudiniOP1 -lHoudiniSIM -lHoudiniGEO -lHoudiniPRM -lHoudiniUT"
	elif forge.fe_os == "FE_WIN32" or forge.fe_os == "FE_WIN64":
		dll.linkmap["houdinilibs"] = "TODO"

	forge.deps( ["fexHoudiniDLLib"], deplibs )

	houdini_path = os.getenv('HT')
	if houdini_path:
		module.includemap = { 'houdini' : os.path.join(houdini_path,'include') }

	module.Module('test')

def auto(module):
	test_file = """
#include <SOP/SOP_API.h>

int main(void)
{
	return(0);
}
    \n"""

	houdini_path = os.getenv('HT')
	if houdini_path:
		forge.includemap['houdini'] = os.path.join(houdini_path,'include')

	return forge.cctest(test_file)
