/*	Copyright (C) 2003-2010 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

#include "operator/operator.pmh"

namespace fe
{

void SurfaceCurveOp::initialize(void)
{
	catalog<I32>("Segments")=4;
	catalog<I32>("Segments","low")=1;
	catalog<I32>("Segments","high")=8;
	catalog<I32>("Segments","min")=1;
	catalog<I32>("Segments","max")=100;

	catalog<Real>("Length")=1.0;
	catalog<Real>("Length","low")=0.0;
	catalog<Real>("Length","high")=10.0;
	catalog<Real>("Length","min")=0.0;
	catalog<Real>("Length","max")=100.0;
}

void SurfaceCurveOp::handle(Record& a_rSignal)
{
//	feLog("SurfaceCurveOp::handle\n");

	sp<DrawI> spDrawI=drawI(a_rSignal);
	if(!spDrawI.isValid())
	{
		feLog("OperatorDrawTest::handle DrawI invalid\n");
		return;
	}

	const I32 segments=catalog<I32>("Segments");
	const Real length=catalog<Real>("Length");
	const Real scale=length/(Real)segments;

	SpatialVector* pCurve=new SpatialVector[segments+1];
	for(I32 index=0;index<=segments;index++)
	{
		set(pCurve[index],0.001*index*index,index*scale);
		feLog("SurfaceCurveOp::handle index %d/%d: %s\n",
				index,segments+1,c_print(pCurve[index]));
	}

	spDrawI->drawLines(pCurve,NULL,segments+1,DrawI::e_strip,false,NULL);
	delete pCurve;

//	feLog("SurfaceCurveOp::handle done\n");
}

} // namespace
