#include "operator/operator.pmh"

#define FE_SSP_DEBUG	FALSE

namespace fe
{

void SurfaceSummaryOp::initialize(void)
{
	catalog< sp<Component> >("Input Surface");

	catalog<I32>("pointCount");
	catalog<I32>("pointCount","max")=INT_MAX;
	catalog<String>("pointCount","IO")="output";
	catalog<String>("pointCount","label")="Points (output)";

	catalog<I32>("primitiveCount");
	catalog<I32>("primitiveCount","max")=INT_MAX;
	catalog<String>("primitiveCount","IO")="output";
	catalog<String>("primitiveCount","label")="Primitives (output)";

	catalog<String>("probeAttribute")="P";
	catalog<String>("probeAttribute","label")="Probe Attribute";

	catalog<String>("probeClass")="Point";
	catalog<String>("probeClass","choice:0")="Point";
	catalog<String>("probeClass","choice:1")="Primitives";
	catalog<String>("probeClass","label")="Probe Class";

	catalog<String>("probeType")="Real";
	catalog<String>("probeType","choice:0")="Real";
	catalog<String>("probeType","choice:1")="Int";
	catalog<String>("probeType","label")="Probe Type";

	catalog<I32>("probeIndex")=0;
	catalog<I32>("probeIndex","high")=100;
	catalog<I32>("probeIndex","max")=INT_MAX;
	catalog<String>("probeIndex","label")="Probe Index";

	catalog<Real>("probeValue")=0.0;
	catalog<Real>("probeValue","min")= -1e9;
	catalog<Real>("probeValue","max")=1e9;
	catalog<String>("probeValue","IO")="output";
	catalog<String>("probeValue","label")="Probe (output)";
}

void SurfaceSummaryOp::handle(Record& a_rSignal)
{
#if FE_SSP_DEBUG
	feLog("SurfaceSummaryOp::handle\n");
#endif

	I32& pointCount=catalog<I32>("pointCount");
	I32& primitiveCount=catalog<I32>("primitiveCount");

	const String probeClass=catalog<String>("probeClass");
	const String probeType=catalog<String>("probeType");
	const String probeAttribute=catalog<String>("probeAttribute");
	const I32 probeIndex=catalog<I32>("probeIndex");
	Real& probeValue=catalog<Real>("probeValue");

	sp<SurfaceAccessibleI> spSurfaceAccessibleI=surfaceOutput(a_rSignal);
	FEASSERT(spSurfaceAccessibleI.isValid());

	sp<SurfaceAccessorI> spAccessor=spSurfaceAccessibleI->accessor(
			SurfaceAccessibleI::e_point,SurfaceAccessibleI::e_generic);
	FEASSERT(spAccessor.isValid());

	pointCount=spAccessor->count();

#if FE_SSP_DEBUG
	feLog("SurfaceSummaryOp::handle pointCount=%d\n",pointCount);
#endif

	spAccessor=spSurfaceAccessibleI->accessor(
			SurfaceAccessibleI::e_primitive,SurfaceAccessibleI::e_generic);
	FEASSERT(spAccessor.isValid());

	primitiveCount=spAccessor->count();

#if FE_SSP_DEBUG
	feLog("SurfaceSummaryOp::handle primitiveCount=%d\n",primitiveCount);
#endif

	bool probePoint=(probeClass=="Point");
	bool probeReal=(probeType=="Real");
	I32 probeMax=probePoint? pointCount: primitiveCount;
	if(probeIndex<probeMax)
	{
		sp<SurfaceAccessibleI> spSurfaceInput=
				catalog< sp<Component> >("Input Surface");
		FEASSERT(spSurfaceInput.isValid());

		spAccessor=spSurfaceInput->accessor(probePoint?
				SurfaceAccessibleI::e_point: SurfaceAccessibleI::e_primitive,
				probeAttribute);
		if(spAccessor.isValid())
		{
			// TODO auto check type
			if(probeReal)
			{
				probeValue=spAccessor->real(probeIndex,0);
			}
			else
			{
				probeValue=spAccessor->integer(probeIndex,0);
			}
#if FE_SSP_DEBUG
			feLog("SurfaceSummaryOp::handle probeValue=%.6G\n",
					probeValue);
#endif
		}
	}
	else
	{
		probeValue= -1.0;
	}

#if FE_SSP_DEBUG
	feLog("SurfaceSummaryOp::handle done\n");
#endif
}

} // namespace
